package com.gyee.wisdom.dao.golden;

import com.golden.api.impl.GoldenServerImpl;
import com.gyee.wisdom.common.exception.WisdomException;
import com.rtdb.service.impl.ServerImpl;
import org.springframework.stereotype.Component;

import java.util.Enumeration;
import java.util.Vector;

@Component
public class GoldenConnectionPool {

    private String serverIp = "127.0.0.1"; // 服务器IP
    private short serverPort = 10010; // 服务器端口
    private String userName = "sa";
    private String password = "golden";
    private int defaultConnections = 1; // 连接池的初始大小
    private int incrementalConnections = 1;// 连接池自动增加的大小
    private int maxConnections = 10; // 连接池最大的大小
    private Vector connections = null; // 存放连接池中连接的向量 , 初始时为 null


    public void  init(String serverIp, short serverPort, String userName, String password, int defaultConnections, int incrementalConnections, int maxConnections) {
        this.serverIp = serverIp;
        this.serverPort = serverPort;
        this.userName = userName;
        this.password = password;
        this.defaultConnections = defaultConnections;
        this.incrementalConnections = incrementalConnections;
        this.maxConnections = maxConnections;
    }

    /**
     * 创建一个连接池，连接池中的可用连接的数量采用类成员 defaultConnections 中设置的值
     */
    private synchronized void createPool() throws Exception {
        // 确保连接池没有创建
        // 如果连接池己经创建了，保存连接的向量 connections 不会为空
        if (connections != null) {
            return; // 如果己经创建，则返回
        }
        // 创建保存连接的向量 , 初始时有 0 个元素
        connections = new Vector();
        // 根据 defaultConnections 中设置的值，创建连接。
        createConnections(this.defaultConnections);
        System.out.println(" Golden连接池创建成功！ ");
    }

    /**
     * 创建由 numConnections 指定数目的Golden连接 , 并把这些连接
     * 放入 connections 向量中
     * @param numConnections  要创建的Golden连接的数目
     */
    private void createConnections(int numConnections) throws Exception {
        // 循环创建指定数目的Golden连接
        for (int x = 0; x < numConnections; x++) {
            // 是否连接池中的Golden连接的数量己经达到最大？最大值由类成员 maxConnections
            // 指出，如果 maxConnections 为 0 或负数，表示连接数量没有限制。
            // 如果连接数己经达到最大，即退出。
            if (this.maxConnections > 0  && this.connections.size() >= this.maxConnections) {
                break;
            }

            // add a new PooledConnection object to connections vector
            // 增加一个连接到连接池中（向量 connections 中）
            try {
                connections.addElement(new PooledConnection(newConnection()));

            } catch (WisdomException e) {
                System.out.println(" 创建Golden连接失败！ " + e.getMessage());
                throw new WisdomException();
            }
            System.out.println(" Golden连接己创建 ......");
        }
    }

    /**
     * 创建一个新的Golden连接并返回它
     * @return 返回一个新创建的Golden连接
     */
    private ServerImpl newConnection() throws Exception {
        // 创建一个Golden连接
//        String file = RtdbServerImpl.class.getProtectionDomain().getCodeSource().getLocation().getFile();
//        System.out.println("fffffffffffffffffffff = " + file);
//        if (file.endsWith(".jar")) {
//            System.out.println("对对对对对对对对对对对对对对");
//        } else {
//            System.out.println("错错错错错错错错错错错错错错");
//        }

        ServerImpl conn = new ServerImpl(serverIp, serverPort, userName, password);
        return conn; // 返回创建的新的Golden连接
    }

    /**
     * 通过调用 getFreeConnection() 函数返回一个可用的Golden连接 ,
     * 如果当前没有可用的Golden连接，并且更多的Golden连接不能创建（如连接池大小的限制），此函数等待一会再尝试获取。
     *
     * @return 返回一个可用的Golden连接对象
     */
    public synchronized ServerImpl getConnection() throws Exception {
        // 确保连接池己被创建
        if (connections == null) {
            createPool(); // 连接池还没创建，则返回 null
        }

        ServerImpl conn = getFreeConnection(); // 获得一个可用的Golden连接
        // 如果目前没有可以使用的连接，即所有的连接都在使用中
        while (conn == null) {
            // 等一会再试
            wait(250);
            conn = getFreeConnection(); // 重新再试，直到获得可用的连接，如果
            // getFreeConnection() 返回的为 null
            // 则表明创建一批连接后也不可获得可用连接
        }
        return conn;// 返回获得的可用的连接
    }

    /**
     * 本函数从连接池向量 connections 中返回一个可用的的Golden连接，如果
     * 当前没有可用的Golden连接，本函数则根据 incrementalConnections 设置
     * 的值创建几个Golden连接，并放入连接池中。
     * 如果创建后，所有的连接仍都在使用中，则返回 null
     * @return 返回一个可用的Golden连接
     */
    private ServerImpl getFreeConnection() throws Exception {
        // 从连接池中获得一个可用的Golden连接
        ServerImpl conn = findFreeConnection();
        if (conn == null) {

            // 如果目前连接池中没有可用的连接
            // 创建一些连接
            createConnections(incrementalConnections);
            // 重新从池中查找是否有可用连接
            conn = findFreeConnection();
            if (conn == null) {
                // 如果创建连接后仍获得不到可用的连接，则返回 null
                return null;
            }
        }
        return conn;
    }

    /**
     * 查找连接池中所有的连接，查找一个可用的Golden连接，
     * 如果没有可用的连接，返回 null
     * @return 返回一个可用的Golden连接
     */
    private ServerImpl findFreeConnection() throws Exception {
        ServerImpl conn = null;
        PooledConnection pConn = null;
        // 获得连接池向量中所有的对象
        Enumeration enumerate = connections.elements();

        // 遍历所有的对象，看是否有可用的连接
        while (enumerate.hasMoreElements()) {
            pConn = (PooledConnection) enumerate.nextElement();
            if (!pConn.isBusy()) {
                // 如果此对象不忙，则获得它的Golden连接并把它设为忙
                conn = pConn.getConnection();
                pConn.setBusy(true);
                break; // 己经找到一个可用的连接，退出
            }
        }
        return conn;// 返回找到到的可用连接
    }


    /**
     * 此函数返回一个Golden连接到连接池中，并把此连接置为空闲。
     * 所有使用连接池获得的Golden连接均应在不使用此连接时返回它。
     */
    public void returnConnection(ServerImpl conn) {
        // 确保连接池存在，如果连接没有创建（不存在），直接返回
        if (connections == null) {
            System.out.println(" 连接池不存在，无法返回此连接到连接池中 !");
            return;
        }

        PooledConnection pConn = null;
        Enumeration enumerate = connections.elements();
        // 遍历连接池中的所有连接，找到这个要返回的连接对象
        while (enumerate.hasMoreElements()) {
            pConn = (PooledConnection) enumerate.nextElement();
            // 先找到连接池中的要返回的连接对象
            if (conn == pConn.getConnection()) {
                // 找到了 , 设置此连接为空闲状态
                pConn.setBusy(false);
                break;
            }
        }
    }

    /**
     * 刷新连接池中所有的连接对象
     */
    public synchronized void refreshConnections() throws Exception {
        // 确保连接池己创新存在
        if (connections == null) {
            System.out.println(" 连接池不存在，无法刷新 !");
            return;
        }

        PooledConnection pConn = null;
        Enumeration enumerate = connections.elements();
        while (enumerate.hasMoreElements()) {

            // 获得一个连接对象
            pConn = (PooledConnection) enumerate.nextElement();
            // 如果对象忙则等 5 秒 ,5 秒后直接刷新
            if (pConn.isBusy()) {
                wait(5000); // 等 5 秒
            }
            // 关闭此连接，用一个新的连接代替它。
            closeConnection(pConn.getConnection());
            pConn.setConnection(newConnection());
            pConn.setBusy(false);
        }
    }

    /**
     * 关闭连接池中所有的连接，并清空连接池。
     */
    private synchronized void closeConnectionPool() throws WisdomException {
        // 确保连接池存在，如果不存在，返回
        if (connections == null) {
            System.out.println(" 连接池不存在，无法关闭 !");
            return;
        }

        PooledConnection pConn = null;
        Enumeration enumerate = connections.elements();
        while (enumerate.hasMoreElements()) {
            pConn = (PooledConnection) enumerate.nextElement();
            // 如果忙，等 5 秒
            if (pConn.isBusy()) {
                wait(5000); // 等 5 秒
            }
            // 5 秒后直接关闭它
            closeConnection(pConn.getConnection());
            // 从连接池向量中删除它
            connections.removeElement(pConn);
        }
        // 置连接池为空
        connections = null;
    }

    /**
     * 关闭一个Golden连接
     */
    private void closeConnection(ServerImpl conn) {
        try {
            conn.close();
        } catch (Exception e) {
            System.out.println(" 关闭Golden连接出错： " + e.getMessage());
        }
    }

    @Override
    public void finalize() {
        try {
            closeConnectionPool();
        } catch (WisdomException e) {
            System.out.println(" 关闭Golden连接池出错： " + e.getMessage());
        }
    }


    /**
     * 使程序等待给定的毫秒数
     */
    private void wait(int mSeconds) {
        try {
            Thread.sleep(mSeconds);

        } catch (InterruptedException e) {
        }
    }

    /**
     * 返回连接池的初始大小
     *
     * @return 初始连接池中可获得的连接数量
     */
    public int getDefaultConnections() {
        return this.defaultConnections;
    }

    /**
     * 设置连接池的初始大小
     */
    public void setDefaultConnections(int defaultConnections) {
        this.defaultConnections = defaultConnections;
    }

    /**
     * 返回连接池自动增加的大小 、
     * @return 连接池自动增加的大小
     */
    public int getIncrementalConnections() {
        return this.incrementalConnections;
    }

    /**
     * 设置连接池自动增加的大小
     */
    public void setIncrementalConnections(int incrementalConnections) {
        this.incrementalConnections = incrementalConnections;
    }

    /**
     * 返回连接池中最大的可用连接数量
     * @return 连接池中最大的可用连接数量
     */
    public int getMaxConnections() {
        return this.maxConnections;
    }

    /**
     * 设置连接池中最大可用的连接数量
     */
    public void setMaxConnections(int maxConnections) {
        this.maxConnections = maxConnections;
    }

    /**
     * 内部使用的用于保存连接池中连接对象的类
     * 此类中有两个成员，一个是Golden的连接，另一个是指示此连接是否
     * 正在使用的标志。
     */
    class PooledConnection {
        ServerImpl connection = null;// Golden连接
        boolean busy = false; // 此连接是否正在使用的标志，默认没有正在使用

        // 构造函数，根据一个 Connection 构告一个 PooledConnection 对象
        public PooledConnection(ServerImpl connection) {
            this.connection = connection;
        }

        // 返回此对象中的连接
        public ServerImpl getConnection() {
            return connection;
        }

        // 设置此对象的，连接
        public void setConnection(ServerImpl connection) {
            this.connection = connection;
        }

        // 获得对象连接是否忙
        public boolean isBusy() {
            return busy;
        }

        // 设置对象的连接正在忙
        public void setBusy(boolean busy) {
            this.busy = busy;
        }

    }

}
